# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

from functools import partial
from typing import Literal, Union, Callable

import numpy as np
import pandas as pd
from sklearn.preprocessing import FunctionTransformer, StandardScaler, OrdinalEncoder
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit
from sklearn.pipeline import Pipeline
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
import numpy.typing as npt

from .globals import translator, models
from .helpers import dispmay_item


def make_num_sanitizer(min_val: np.number, max_val: np.number) -> FunctionTransformer:
    """
    Создаёт трансформер, который заполняет все значения, выходящие за
    указанный диапазон значениями np.nan
    """

    def coerce_num_vals(data: np.ndarray, min_val: np.number, max_val: np.number):
        data = data.astype("float")
        data[data < min_val] = np.nan
        data[data > max_val] = np.nan
        return data

    return FunctionTransformer(
        func=coerce_num_vals,
        validate=True,
        kw_args={"min_val": min_val, "max_val": max_val},
        feature_names_out="one-to-one",
    )


def make_typo_corrector(correct_vals: list[str]) -> FunctionTransformer:
    """
    Простой корректировщик опечаток.

    Считает расстояния Хэмминга между словами и корректирует одиночную опечатку.
    """

    def correct_typo(word: Union[str, float], correct_vals: list[str]):
        if isinstance(word, (int, float)):
            return np.nan

        hamming_vals = {}
        for value in correct_vals:
            if len(word) != len(value):
                hamming_vals[value] = np.inf
            else:
                hamming_vals[value] = sum(ch1 != ch2 for ch1, ch2 in zip(word, value))

        min_dist = min(hamming_vals.values())
        if min_dist == 0:
            return word
        if min_dist == 1:
            for key, val in hamming_vals.items():
                if val == 1:
                    return key
        return np.nan

    vect = np.vectorize(
        partial(correct_typo, correct_vals=correct_vals), otypes=[object]
    )
    return FunctionTransformer(func=vect, feature_names_out="one-to-one")


def make_num_processor(min_val: np.number, max_val: np.number) -> Pipeline:
    """
    Создаёт препроцессор для численных данных.
    """
    return Pipeline(
        [
            ("sanitizer", make_num_sanitizer(min_val, max_val)),
            ("imputer", IterativeImputer()),
            ("num_scaler", StandardScaler()),
        ]
    )


# Предвижу, что возможно эта чусть будет помечена как ошибка во время ревью
# Я сознательно не стал использовать imputer перед кодированием, так как не
# вижу в нём смысла. Пропуски и неизвестные категории будут закодированы
# ординал энкодером как пропуски (np.nan), после чего IterativeImputer заполнит
# их подходящими значениями, поэтому я не понимаю зачем использовать imputer
# 2 раза (до и после энкодера).
def make_ord_processor(categories: list[str]) -> Pipeline:
    """
    Создаёт препроцессор для категориальных данных,
    которые будут кодироваться с помощью OrdinalEncoder.
    """
    return Pipeline(
        [
            ("typo_corrector", make_typo_corrector(categories)),
            (
                "ord_encoder",
                OrdinalEncoder(
                    categories=[categories],
                    handle_unknown="use_encoded_value",
                    unknown_value=np.nan,
                ),
            ),
            ("imputer", IterativeImputer()),
            ("rounder", FunctionTransformer(np.round, feature_names_out="one-to-one")),
        ]
    )


def make_num_ord_processor(min_val: np.number, max_val: np.number) -> Pipeline:
    """
    Создаёт препроцессор для численных данных, которые по сути являются
    категориальными (аналог make_ord_processor).
    """

    return Pipeline(
        [
            ("sanitizer", make_num_sanitizer(min_val, max_val)),
            ("imputer", IterativeImputer()),
            ("rounder", FunctionTransformer(np.round, feature_names_out="one-to-one")),
        ]
    )


def show_search_result(search: GridSearchCV, n_results: int = 10) -> pd.DataFrame:
    """
    Возвращает датафрйем с основными результатами поиска по сетке.
    """
    df = pd.DataFrame(search.cv_results_)
    report_cols = [
        "rank_test_score",
        "mean_test_score",
        "mean_fit_time",
        "mean_score_time",
    ]
    report_cols.extend([col for col in df.columns if col.startswith("param_")])
    return df[report_cols].sort_values(by="rank_test_score").head(n_results).round(4)


def evaluate_params(grid: GridSearchCV):
    """
    Выводит максимальное значение метрики качества для значений гиперпараметров
    """
    df = pd.DataFrame(grid.cv_results_)
    params = [col for col in df.columns if col.startswith("param_")]

    for param in params:
        dispmay_item(
            df[[param, "mean_test_score", "rank_test_score"]]
            .groupby(param, sort=False)
            .agg({"mean_test_score": "max", "rank_test_score": "min"})
            .sort_values(by="mean_test_score", ascending=False)
            .head(10)
            .round(4)
        )


def grid_search(
    pipeline: Pipeline,
    grid: Union[list, dict],
    X_train: pd.DataFrame,
    y_train: Union[pd.DataFrame, np.ndarray],
    model: str,
    scoring: Union[Callable, str] = 'roc_auc',
    cv_method: Union[TimeSeriesSplit, int, None] = None,
    n_jobs: int = -1,
) -> None:
    """
    Осуществляет поиск по сетке, выводит и сохраняет результаты.
    """


    gs_kwargs = {
        "scoring": scoring,
        "cv": cv_method if cv_method else 5,
        "n_jobs": n_jobs,
        "verbose": 1,
    }
    grid_search_result = GridSearchCV(pipeline, grid, **gs_kwargs).fit(X_train, y_train)
    result = show_search_result(grid_search_result)
    dispmay_item(result)
    cols = ["mean_test_score", "mean_fit_time", "mean_score_time"]
    best_result = result.loc[result["rank_test_score"] == 1, cols].iloc[0].copy()
    models.loc[model, cols] = best_result
    translator.update(
        {
            "mean_test_score": "Ср. результат",
            "mean_fit_time": "Ср. длительность обучения",
            "mean_score_time": "Ср. длительность предсказания",
            "model": "Модель",
        }
    )
    models.at[model, "model"] = grid_search_result.best_estimator_
    models.sort_values("mean_test_score", ascending=False, inplace=True)



# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

from sklearn.model_selection import GridSearchCV
import pandas as pd

translator = {}
models: pd.DataFrame = pd.DataFrame()
scores: dict[str, float] = {}


# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

""" import models as models
import plotting as plotting
import data_handling as data_handling
from .globals import * """

#__all__ = ["models", "plotting", "data_handling"]


# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

def is_jupyter():
    "Check environment"
    try:
        # Check if we're in IPython/Jupyter
        from IPython import get_ipython
        if get_ipython() is None:
            return False
        # Check if it's a Jupyter notebook/lab
        if 'IPKernelApp' in get_ipython().config:
            return True
        # Alternative check
        if hasattr(get_ipython(), 'kernel'):
            return True
        return False
    except ImportError:
        return False
    
def dispmay_item(item_to_display: object) -> None:
    if is_jupyter():
        display(item_to_display)
    else:
        print(item_to_display)


# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Sequence, Optional, cast
import math

import pandas as pd
import numpy as np
import numpy.typing as npt
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.ticker as ticker

from .data_handling import get_num_cols

def plot_hist(
    df: pd.DataFrame,
    col: str,
    ax: Optional[Axes] = None,
    hue: Optional[str] = None,
    annotate: bool = True,
    kde: bool = False,
    normalize: bool = False,
    **kwargs,
) -> Axes:
    """
    Строит гистрограмму c опциональным разбиением на категории.

    Параметры
    ----------
    - df: Датафрейм с данными
    - col: Название колонки, содержащей данные для построения гистограммы
    - ax: оси, на которых нужно строить гистограмму
    - hue: Наименование опционального столбца для разделения данных
    - annotate: Добавлят подписи к оси Х и заголовок графика
    - normalize: Нормировать гистограмму
    """
    if normalize:
        stat = "probability"
    else:
        stat = "count"

    ax = sns.histplot(
        df,
        x=col,
        hue=hue,
        discrete=pd.api.types.is_integer(df[col]),
        kde=kde,
        kde_kws={"bw_method": 0.6} if kde else None,
        edgecolor="white",
        stat=stat,
        common_norm=False,
        ax=ax,
        **kwargs,
    )
    if normalize:
        ax.set_ylabel("Плотность распределения")
    else:
        ax.set_ylabel("Количество")
    if annotate:
        ax.set_title(f'Гистограмма распределение признака\n"{col}"')
        ax.set_xlabel(col)
    else:
        ax.set_xlabel("")
    return ax


def plot_box(
    df: pd.DataFrame,
    col: str,
    ax: Optional[Axes] = None,
    hue: Optional[str] = None,
    annotate: bool = True,
    **kwargs,
):
    """
    Строит боксплот c опциональным разбиением на категории.

    Параметры
    ----------
    - df: Датафрейм с данными
    - col: Название колонки, содержащей данные для построения боксплота
    - ax: оси, на которых нужно строить боксплот
    - hue: Наименование опционального столбца для разделения данных
    - annotate: Добавлят подписи к оси Х и заголовок графика
    """
    ax = sns.boxplot(df, y=col, hue=hue, ax=ax, **kwargs)
    if annotate:
        ax.set_title(f'Боксплот распределение признака\n"{col}"')
        ax.set_xlabel(col)
    else:
        ax.set_xlabel("")
        ax.set_ylabel("")


def plot_numeric(
    df: pd.DataFrame,
    num_cols: Optional[list[str]] = None,
    title: str = "",
    hue: Optional[str] = None,
    normalize: bool = False,
    kde: bool = True,
    ncols: int = 2,
    scale: float = 2.5,
    **kwargs,
) -> None:
    """
    Функция для построения гистограмм и боксплотов по числовым признакам.

    Параметры:
    ----------
    - df: DataFrame с данными.
    - num_cols: Список числовых признаков для анализа.
    - title: Заголовок для графиков.
    - hue: Опциональный столбец для разделения данных по категориям.
    - normalize: Нормировать гистограмму
    - ncols: Количество столбцов для размещения графиков.
    - scale: Скейлинг фактор, определяющий размер
    """

    if num_cols is None:
        num_cols = get_num_cols(df)
    if len(num_cols) == 1:
        ncols = 1
    nrows = math.ceil(len(num_cols) / ncols)
    if ncols > 1:
        fig = plt.figure(
            layout="constrained", figsize=(scale * ncols * 2, scale * nrows)
        )
    else:
        fig = plt.figure(layout="constrained", figsize=(scale * 4, scale * nrows * 2))
    fig.suptitle(title, fontsize=16)
    subfigs = fig.subfigures(ncols=ncols, nrows=nrows, squeeze=True)
    if isinstance(subfigs, np.ndarray):
        subfigs = cast(list[Figure], subfigs.flatten())
    else:
        subfigs = [subfigs]
    for i, col in enumerate(num_cols):
        axs = subfigs[i].subplots(1, 2)
        if len(title) == 0 and len(num_cols) != 0:
            subfigs[i].suptitle(f"Распределение {col}")
        plot_hist(
            df, col, axs[0], hue, annotate=False, normalize=normalize, kde=kde, **kwargs
        )
        plot_box(df, col, axs[1], hue, annotate=False, legend=False, gap=0.2, **kwargs)


def plot_cats(
    df: pd.DataFrame,
    cat_cols: Optional[list[str]] = None,
    hue: Optional[str] = None,
    title: str = "",
    ncols: int = 2,
    max_cats: int = 10,
    max_cats_alias: str = "all_other",
    **kwargs,
) -> None:
    """
    Функция для построения графиков распределения категориальных признаков.

    Параметры:
    ----------
    - df: DataFrame с данными.
    - cat_cols: Список категориальных признаков для анализа.
    - hue: Название категориального признака, по которому будут группирвоаться значения.
    - title: Заголовок для графиков.
    - ncols: Количество столбцов для размещения графиков.
    - max_cats: Максимальное количество категорий. Более мелкие категории будут
    объединены в одну категорию.
    - max_cats_alias: Назкание категории, которая обхединяет более мелкие категории.
    """

    if cat_cols is None:
        cat_cols = df.select_dtypes([object, "category"]).columns.to_list()
    if len(cat_cols) == 0:
        return
    scale = 3
    nrows = math.ceil(len(cat_cols) / ncols)
    fig, axs = plt.subplots(nrows, ncols, figsize=(scale * ncols * 2, scale * nrows))
    fig.suptitle(title, fontsize=16)
    fig.set_layout_engine("constrained")
    try:
        axs = cast(Sequence[Axes], axs.flatten())
    except AttributeError:
        axs = cast(Sequence[Axes], [axs])
    for i, col in enumerate(cat_cols):

        if hue is not None:
            data = (
                df.groupby(hue)[col]
                .value_counts(normalize=True, ascending=False)
                .reset_index()
            )
            if data.shape[0] > max_cats:
                data.iloc[max_cats] = data.iloc[max_cats:].sum()
                data = data.iloc[: max_cats + 1]
                data = data.rename(index={data.index[max_cats]: max_cats_alias})
            sns.barplot(
                data,
                x=data.columns.difference([col, hue])[0],
                y=col,
                hue=hue,
                ax=axs[i],
                orient="y",
                **kwargs,
            )
            for container in axs[i].containers:
                axs[i].bar_label(
                    container, label_type="edge", padding=3, fmt=lambda x: f"{x:.0%}"  # type: ignore
                )
                x_min, x_max = axs[i].get_xlim()
                axs[i].set_xlim(x_min, x_max + (x_max - x_min) * 0.05)
        else:
            data = df[col].value_counts(ascending=False)
            if data.shape[0] > max_cats:
                data.iloc[max_cats] = data.iloc[max_cats:].sum()
                data = data.iloc[: max_cats + 1]
                data = data.rename(index={data.index[max_cats]: max_cats_alias})
            sns.barplot(data, ax=axs[i], orient="h", color="#8ebad9", **kwargs)  # type: ignore
            axs[i].bar_label(
                axs[i].containers[0],  # type: ignore
                label_type="edge",
                padding=5,
                fmt=lambda x: f"{int(x)} ({x/data.sum():.0%})",
            )
            x_min, x_max = axs[i].get_xlim()
            axs[i].set_xlim(x_min, x_max + (x_max - x_min) * 0.25)
        axs[i].set_title(f"Распределение {col}", fontsize=14)
        axs[i].set_xlabel("")
        axs[i].set_ylabel("")
        axs[i].xaxis.set_major_formatter(ticker.StrMethodFormatter("{x:,.0f}"))
    # Удалим лишние оси
    for i in range(len(cat_cols), nrows * ncols):
        axs[i].remove()

def show_residues(y_true: npt.ArrayLike, y_pred: npt.ArrayLike, title: str = "", **kwargs):
    """
    Строит графики распределения остатков

    Параметры
    --------
    - y_true: Истинные значения
    - y_pred: Предсказанные значения
    - title: Опциональный общий заголовок для всего изображения
    """
    data = pd.DataFrame()
    data["истина"] = y_true
    data["предсказание"] = y_pred
    data["остатки"] = data["предсказание"] - data["истина"]
    fig, axs = plt.subplots(1, 2, figsize=(10, 4))
    if len(title):
        fig.suptitle(title)
    plot_hist(data, col="остатки", ax=axs[0], **kwargs)
    sns.scatterplot(data, x="истина", y="остатки", ax=axs[1], **kwargs)
    _ = axs[1].set_title(
        "Диаграмма рассеяния остатков\nв зависимости от истинных значений"
    )
    fig.tight_layout()


# SPDX-FileCopyrightText: 2025-present Anton Popov <a.popov.fizteh@gmail.com>
#
# SPDX-License-Identifier: MIT

from typing import Optional, Union, cast

import numpy as np
import pandas as pd
import phik

from .helpers import dispmay_item
from .globals import translator

__all__ = ["show_df", "get_num_cols", "get_cat_cols", "print_unique_cat_vals", "highest_corrs"]


def show_df(df: pd.DataFrame, n: int = 5) -> None:
    """
    Выводит основную информацию о датафрейме.

    Параметры:
    ----------
    - df: DataFrame с данными.
    - n: Количество строк для отображения.
    """


    print(f"Первые {n} строк:")
    dispmay_item(df.head(n))
    try:
        temp = df.describe(include=[np.number])
        print("\nСтатистика по числовым столбцам:")
        dispmay_item(temp)
    except ValueError:
        print("\nНет числовых столбцов для отображения.")
    try:
        temp = df.describe(include=[object])
        print("\nСтатистика по категориальным столбцам:")
        dispmay_item(temp)
    except ValueError:
        print("\nНет категориальных столбцов для отображения.")
    print("\nИнформация о датафрейме:")
    df.info()


def get_num_cols(
    df: pd.DataFrame, exclude_cols: Optional[list[str]] = None
) -> list[str]:
    """
    Возвращает список числовых столбцов в датафрейме, исключая указанные колонки.

    Параметры
    ----------
    - df: Датафрейм с данными
    - exclude_cols: Список колонок, которые нужно исключить из результата
    """
    if exclude_cols is None:
        exclude_cols = []
    return list(df.drop(exclude_cols, axis=1).select_dtypes(np.number).columns)


def get_cat_cols(
    df: pd.DataFrame, exclude_cols: Optional[list[str]] = None
) -> list[str]:
    """
    Возвращает список категориальных столбцов в датафрейме, исключая указанные колонки.

    Параметры
    ----------
    - df: Датафрейм с данными
    - exclude_cols: Список колонок, которые нужно исключить из результата
    """
    if exclude_cols is None:
        exclude_cols = []
    return list(
        df.drop(exclude_cols, axis=1).select_dtypes([object, "category"]).columns
    )


def print_unique_cat_vals(
    dfs: Union[dict[str, pd.DataFrame], pd.DataFrame], exlculde: Optional[list[str]]
) -> None:
    """
    Выводит уникальные значения категориальных признаков.

    Может работать с отдельными датафреймами или с наборами (слоаврями) датафреймов.

    Параметры
    ---------
    - dfs: датафрейм или словарь датафреймов
    - exclude: список колонок, которые нужно исключить
    """

    def print_unique(df: pd.DataFrame) -> None:
        for j, col in enumerate(get_cat_cols(df, exclude_cols=exlculde), start=1):
            print(f'\t{j}. Признак "{col}". Уникальные значения:')
            print(f"\t {df[col].unique().tolist()}")

    print("Уникальные значения категориальных признаков:")
    if isinstance(dfs, pd.DataFrame):
        print_unique(dfs)
    else:
        for i, (df_name, df) in enumerate(dfs.items(), start=1):
            print(f"{i}. Датафрейм {df_name}:")
            print_unique(df)


def highest_corrs(
    df: pd.DataFrame,
    cols: Optional[list[str]] = None,
    interval_cols: Optional[list[str]] = None,
    num: int = 10,
) -> pd.DataFrame:
    """
    Функция для вычисления корреляций между парами числовых признаков

    Параметры:
    ----------
    - df: DataFrame с данными.
    - cols: Список числовых признаков для анализа.
    - interval_cols: Список численных признаков
    - num: Количество значений для вывода.
    """
    if cols is None:
        cols = list(df.columns)
    try:
        df = df.rename(translator, axis=1)
    except NameError:
        pass
    corr_matrix = cast(pd.DataFrame, df[cols].phik_matrix(interval_cols=interval_cols))  # type: ignore

    corrs = (
        corr_matrix.pipe(lambda x: x.where(np.triu(np.ones_like(x), k=1).astype(bool)))
        .stack()
        .dropna()
        .sort_values(key=abs, ascending=False)  # type: ignore
        .to_frame()
        .round(2)
        .head(num)
    )
    corrs = corrs.set_axis(["corrs"], axis=1)

    def signigicance(cor: float) -> str:
        cor = abs(cor)
        if cor < 0.25:
            return "отсутствует"
        if cor < 0.5:
            return "слабая"
        if cor < 0.75:
            return "средняя"
        return "сильная"

    corrs["значимость"] = corrs["corrs"].transform(signigicance)
    return corrs
